#!/usr/bin/env python
# -*- coding: utf-8 -*
# doc: https://docs.python.org/3/library/codecs.html#standard-encodings
# doc: http://www.fileformat.info/info/unicode/char/search.htm

import os
import sys
import yaml
import json
import time
import shutil
import string
import random
import sqlite3
import datetime
from optparse import OptionParser
from glob import glob
from slug import slug
from tqdm import tqdm
import jinja2

scriptname = os.path.basename(sys.argv[0])
scriptversion='0.3.0'

''' DEFAULT CONFIG OPTIONS (YAML) '''
def join(loader, node):
    seq = loader.construct_sequence(node)
    return ''.join([str(i) for i in seq])
    
yaml.add_constructor('!join', join)
cfg = yaml.load(open(glob('build_*.yml')[0]), Loader=yaml.Loader)
cfg['site_defaults']['year'] = datetime.date.today().year
default     = cfg['site_defaults']
default_idx = cfg['index_defaults']

''' VARIABLES '''

''' FUNCTIONS '''

''' SETUP ENVIRONMENT '''
def setup(*args):
    ''' CHECK FOR TEMPLATE DIR '''
    if not os.path.isdir(cfg['j2Dir']):
        print('Error: template directory %s missing!' % cfg['j2Dir'])
        sys.exit(1)

    ''' CHECK FOR AND GET JSON DATA '''
    if os.path.isdir(cfg['jsonDir']):
        jsonFiles = glob('%s/*/*.json' % cfg['jsonDir'])
        jsonFiles.sort()
    else:
        print('Error: json directory %s missing!' % cfg['jsonDir'])
        sys.exit(1)

    jint = 1
    jsonDict = {}
    for jf in jsonFiles:
        with open(jf) as jfh:
            jsonDict[jint] = json.loads(''.join(jfh.readlines()))
            jsonDict[jint]['json_file']     = os.path.basename(jf)
            jsonDict[jint]['json_dir']      = os.path.dirname(jf)
            jsonDict[jint]['json_abspath']  = os.path.abspath(jf)
            jsonDict[jint]['json_mod_date'] = time.strftime('%Y-%m-%d', time.localtime(os.path.getmtime(jf)))
            jint += 1

    return jsonDict

''' BUILD INDEX OF ARTICLES '''
def buildIndex(*args):
    ''' Required arguments: [jsonDict] [tags] '''
    jsonDict, tags = args[0], args[1]
    browseIdx = {}
    for obj in jsonDict:
        firstChar = jsonDict[obj]['slug'][0].lower()
        if firstChar not in string.ascii_lowercase:
            firstChar = '0-9'
        if firstChar in browseIdx:
            browseIdx[firstChar][jsonDict[obj]['slug']] = jsonDict[obj]
        else:
            browseIdx[firstChar] = {}
            browseIdx[firstChar][jsonDict[obj]['slug']] = jsonDict[obj]

    keys = list(browseIdx.keys())
    keys.sort()
    keys = tags + keys
    for obj in jsonDict:
        for t in tags:
            if t in jsonDict[obj]['tags']:
                if t in browseIdx:
                    browseIdx[t][jsonDict[obj]['slug']] = jsonDict[obj]
                else:
                    browseIdx[t] = {}
                    browseIdx[t][jsonDict[obj]['slug']] = jsonDict[obj]

    return keys, browseIdx, jsonDict

''' CMD LINE '''
def cmdline():
    usage = "%prog [-h]"
    parser = OptionParser(usage=usage, version='%prog ' + scriptversion)
    options, args = parser.parse_args()
    if len(args) > 0:
        parser.print_usage()
        sys.exit(1)
    return options, args

''' MAIN '''
if __name__ == '__main__':
    try:
        jsonDict = setup()

        ''' PROD -OR- DEV '''
        print('Initializing environment...')
        if default['prod'] and default['prod'] == 'prod':
            default['css'] = default['css']['prod']
            default['css_script'] = default['css_script']['prod']
            default['javascript'] = default['javascript']['prod']
        else:
            default['css'] = default['css']['dev']
            default['css_script'] = default['css_script']['dev']
            default['javascript'] = default['javascript']['dev']

        ''' SLUGS '''
        print('Building slugs...')
        slugDB = []
        for obj in jsonDict:
            title_slug = slug(jsonDict[obj]['title'])
            if title_slug in slugDB:
                title_slug = '%s-%s' % (title_slug, obj)
            jsonDict[obj]['slug'] = title_slug
            slugDB.append(title_slug)
            slugDB.sort()

        ''' INDEX / HOME PAGE '''
        print('Building home page...')
        keys, browseIdx, jsonDict = buildIndex(jsonDict, list(default_idx['idxTags'].keys()))
        env = jinja2.Environment(
            loader = jinja2.FileSystemLoader(cfg['j2Dir'], followlinks=True),
            autoescape = jinja2.select_autoescape(['j2'])
        )
        randomArt = []
        for _ in range(default_idx['homeArtCount']):
            article = random.choice(jsonDict)
            while True:
                if article not in randomArt and article['title'].lower().find('quote') == -1:
                    randomArt.append(article)
                    break
                else:
                    article = random.choice(jsonDict)

        template_html = env.get_template('index.html.j2')
        index_html = template_html.render(default = default, articles = randomArt, bkeys = keys).splitlines()
        final_html = [ line for line in index_html if line.strip() ]
        with open('%s/index.html' % (cfg['siteDir']), 'w') as ihtml:
            for html in final_html:
                print(html, file=ihtml)
        print('Build tasks complete.')

    except KeyboardInterrupt:
        sys.exit(0)

