#!/usr/bin/env python
# -*- coding: utf-8 -*
# doc: https://docs.python.org/3/library/codecs.html#standard-encodings
# doc: http://www.fileformat.info/info/unicode/char/search.htm

import os
import sys
import yaml
import json
import time
import shutil
import string
import random
import sqlite3
import datetime
from optparse import OptionParser
from glob import glob
from slug import slug
from tqdm import tqdm
import jinja2
#
scriptname = os.path.basename(sys.argv[0])
scriptversion='0.3.0'

''' DEFAULT CONFIG OPTIONS (YAML) '''
def join(loader, node):
    seq = loader.construct_sequence(node)
    return ''.join([str(i) for i in seq])

yaml.add_constructor('!join', join)
cfg = yaml.load(open(glob('build_*.yml')[0]), Loader=yaml.Loader)
cfg['site_defaults']['year'] = datetime.date.today().year
default     = cfg['site_defaults']
default_idx = cfg['index_defaults']

''' VARIABLES '''

''' FUNCTIONS '''

''' SETUP ENVIRONMENT '''
def setup(*args):
    ''' CHECK FOR TEMPLATE DIR '''
    if not os.path.isdir(cfg['j2Dir']):
        print('Error: template directory %s missing!' % cfg['j2Dir'])
        sys.exit(1)

    ''' CHECK FOR AND GET JSON DATA '''
    if os.path.isdir(cfg['jsonDir']):
        jsonFiles = glob('%s/*/*.json' % cfg['jsonDir'])
        jsonFiles.sort()
    else:
        print('Error: json directory %s missing!' % cfg['jsonDir'])
        sys.exit(1)

    jint = 1
    jsonDict = {}
    for jf in jsonFiles:
        with open(jf) as jfh:
            jsonDict[jint] = json.loads(''.join(jfh.readlines()))
            jsonDict[jint]['json_file']     = os.path.basename(jf)
            jsonDict[jint]['json_dir']      = os.path.dirname(jf)
            jsonDict[jint]['json_abspath']  = os.path.abspath(jf)
            jsonDict[jint]['json_mod_date'] = time.strftime('%Y-%m-%d', time.localtime(os.path.getmtime(jf)))
            jint += 1

    ''' REMOVE SITE DIR '''
    if os.path.isdir(cfg['siteDir']):
        shutil.rmtree(cfg['siteDir'])

    ''' SETUP SITE DIRS '''
    for sdir in (cfg['articleSite'], cfg['cssSite'], cfg['imgSite'], cfg['jsSite'], cfg['fontSite'], cfg['errSite']):
        if not os.path.isdir(sdir):
            os.makedirs(sdir, exist_ok=True)

    return jsonDict

''' BUILD INDEX OF ARTICLES '''
def buildIndex(*args):
    ''' Required arguments: [jsonDict] [tags] '''
    jsonDict, tags = args[0], args[1]
    browseIdx = {}
    for obj in jsonDict:
        firstChar = jsonDict[obj]['slug'][0].lower()
        if firstChar not in string.ascii_lowercase:
            firstChar = '0-9'
        if firstChar in browseIdx:
            browseIdx[firstChar][jsonDict[obj]['slug']] = jsonDict[obj]
        else:
            browseIdx[firstChar] = {}
            browseIdx[firstChar][jsonDict[obj]['slug']] = jsonDict[obj]

    keys = list(browseIdx.keys())
    keys.sort()
    keys = tags + keys
    for obj in jsonDict:
        for t in tags:
            if t in jsonDict[obj]['tags']:
                if t in browseIdx:
                    browseIdx[t][jsonDict[obj]['slug']] = jsonDict[obj]
                else:
                    browseIdx[t] = {}
                    browseIdx[t][jsonDict[obj]['slug']] = jsonDict[obj]

    return keys, browseIdx, jsonDict

''' CMD LINE '''
def cmdline():
    usage = "%prog [-h]"
    parser = OptionParser(usage=usage, version='%prog ' + scriptversion)
    options, args = parser.parse_args()
    if len(args) > 0:
        parser.print_usage()
        sys.exit(1)
    return options, args

''' MAIN '''
if __name__ == '__main__':
    try:
        jsonDict = setup()

        ''' PROD -OR- DEV '''
        print('Initializing environment...')
        if default['prod'] and default['prod'] == 'prod':
            default['css'] = default['css']['prod']
            default['css_script'] = default['css_script']['prod']
            default['javascript'] = default['javascript']['prod']
        else:
            default['css'] = default['css']['dev']
            default['css_script'] = default['css_script']['dev']
            default['javascript'] = default['javascript']['dev']

        ''' SLUGS '''
        print('Building slugs...')
        slugDB = []
        for obj in jsonDict:
            title_slug = slug(jsonDict[obj]['title'])
            if title_slug in slugDB:
                title_slug = '%s-%s' % (title_slug, obj)
            jsonDict[obj]['slug'] = title_slug
            slugDB.append(title_slug)
            slugDB.sort()

        ''' ARTICLES '''
        keys, browseIdx, jsonDict = buildIndex(jsonDict, list(default_idx['idxTags'].keys()))
        print('Building articles...')
        for obj in tqdm(jsonDict):
            slugIdx = slugDB.index(jsonDict[obj]['slug'])
            if slugIdx == 0:
                aprev = jsonDict[[obj for obj in jsonDict if jsonDict[obj]['slug'] == slugDB[-1]][0]]
                anext = jsonDict[[obj for obj in jsonDict if jsonDict[obj]['slug'] == slugDB[slugIdx + 1]][0]]
            elif slugIdx == (len(slugDB) - 1):
                aprev = jsonDict[[obj for obj in jsonDict if jsonDict[obj]['slug'] == slugDB[slugIdx - 1]][0]]
                anext = jsonDict[[obj for obj in jsonDict if jsonDict[obj]['slug'] == slugDB[0]][0]]
            else:
                aprev = jsonDict[[obj for obj in jsonDict if jsonDict[obj]['slug'] == slugDB[slugIdx - 1]][0]]
                anext = jsonDict[[obj for obj in jsonDict if jsonDict[obj]['slug'] == slugDB[slugIdx + 1]][0]]
            env = jinja2.Environment(
                loader = jinja2.FileSystemLoader(cfg['j2Dir'], followlinks=True),
                autoescape = jinja2.select_autoescape(['j2'])
            )
            template_html = env.get_template('article.html.j2')
            article_html = template_html.render(default = default, article = jsonDict[obj], anext = anext, aprev = aprev, bkeys = keys).splitlines()
            final_html = [ line for line in article_html if line.strip() ]
            with open('%s/%s.html' % (cfg['articleSite'], jsonDict[obj]['slug']), 'w') as ahtml:
                for html in final_html:
                    print(html, file=ahtml)

        ''' BROWSE INDEXES [0-9, A-Z] '''
        print('Building browse indexes...')
        for key in tqdm(keys):
            template_html = env.get_template('browse.html.j2')
            fcKeys = list(browseIdx[key].keys())
            fcKeys.sort()
            browse_html = template_html.render(default = default, bidx = browseIdx, bkeys = keys, fckeys = fcKeys, gkey = key, tags = default_idx['idxTags']).splitlines()
            final_html = [ line for line in browse_html if line.strip() ]
            with open('%s/%s.html' % (cfg['siteDir'], key), 'w') as bhtml:
                for html in final_html:
                    print(html, file=bhtml)

        with open('%s_browse.json' % cfg['baseName'], 'w') as bjson:
            json.dump(keys, bjson)

        ''' INDEX / HOME PAGE '''
        print('Building home page...')
        randomArt = []
        for _ in range(default_idx['homeArtCount']):
            article = random.choice(jsonDict)
            while True:
                if article not in randomArt and article['title'].lower().find('quote') == -1:
                    randomArt.append(article)
                    break
                else:
                    article = random.choice(jsonDict)

        template_html = env.get_template('index.html.j2')
        index_html = template_html.render(default = default, articles = randomArt, bkeys = keys).splitlines()
        final_html = [ line for line in index_html if line.strip() ]
        with open('%s/index.html' % (cfg['siteDir']), 'w') as ihtml:
            for html in final_html:
                print(html, file=ihtml)

        ''' SITEMAP.XML '''
        print('Building sitemap...')
        template_xml = env.get_template('sitemap.xml.j2')
        sitemap_xml = template_xml.render(default = default, articles = jsonDict).splitlines()
        final_xml = [ line for line in sitemap_xml if line.strip() ]
        with open('%s/sitemap.xml' % (cfg['siteDir']), 'w') as smxml:
            for xml in final_xml:
                print(xml, file=smxml)

        ''' CSS '''
        print('Building style sheets...')
        template_css = env.get_template('site.css.j2')
        site_css = template_css.render(default = default).splitlines()
        final_css = [ line for line in site_css if line.strip() ]
        with open('%s/%s.css' % (cfg['cssDir'], cfg['baseName']), 'w') as fcss:
            for css in final_css:
                print(css, file=fcss)

        print('Building css path...')
        for css in default['css'] + default['css_script']:
            if os.path.isdir(cfg['cssSite']) and os.path.isfile('%s/%s' % (cfg['cssDir'], css)):
                shutil.copy2('%s/%s' % (cfg['cssDir'], css), cfg['cssSite'])
            else:
                print('Error: CSS file %s missing! Skipping.' % ('%s/%s' % (cfg['cssDir'], css)))

        ''' JAVASCRIPT '''
        if not default['prod']:
            print('Building js path...')
            for js in default['javascript']:
                if os.path.isdir(cfg['jsSite']) and os.path.isfile('%s/%s' % (cfg['jsDir'], js)):
                    shutil.copy2('%s/%s' % (cfg['jsDir'], js), cfg['jsSite'])
                else:
                    print('Error: JS file %s missing! Skipping.' % ('%s/%s' % (cfg['jsDir'], js)))

        ''' ERROR '''
        print('Building error pages...')
        template_html = env.get_template('error.html.j2')
        for erc in tqdm(cfg['errorCode']):
            error_html = template_html.render(default = default, bkeys = keys, error_code=erc, ecodes = cfg['errorCode']).splitlines()
            final_html = [ line for line in error_html if line.strip() ]
            with open('%s/%s.html' % (cfg['errSite'], erc), 'w') as ehtml:
                for html in final_html:
                    print(html, file=ehtml)

        ''' IMG '''
        print('Building image path...')
        if os.path.isdir(cfg['imgSite']) and os.path.isdir(cfg['imgDir']):
            for img in glob('%s/*.svg' % cfg['imgDir']):
                shutil.copy2(img, cfg['imgSite'])
        else:
            print('Error: IMAGE files in %s missing! Skipping.' % cfg['imgDir'])

        ''' FONT '''
        print('Building font path...')
        if os.path.isdir(cfg['fontSite']) and os.path.isdir(cfg['fontDir']):
            for font in glob('%s/*.ttf' % cfg['fontDir']):
                shutil.copy2(font, cfg['fontSite'])
        else:
            print('Error: FONT files in %s missing! Skipping.' % cfg['fontDir'])

        ''' FAVICON '''
        print('Building favicon path...')
        if os.path.isfile(default['favicon']):
            shutil.copy2(default['favicon'], cfg['siteDir'])

        ''' NGINX CONF '''
        print('Building nginx configuration...')
        template_conf = env.get_template('nginx.conf.j2')
        site_conf = template_conf.render(default = default, conf = cfg['site_server'], ecodes = cfg['errorCode']).splitlines()
        final_conf = [ line for line in site_conf if line.strip() ]
        with open('%s.conf' % (cfg['baseName']), 'w') as sconf:
            for conf in final_conf:
                print(conf, file=sconf)

        ''' CONF FILES (robots.txt, etc..) '''
        print('Building conf files...')
        if os.path.isdir(cfg['confDir']):
            for conf in ('%s/robots.txt' % cfg['confDir'],):
                shutil.copy2(conf, cfg['siteDir'])
        else:
            print('Error: CONF files in %s missing! Skipping.' % cfg['confDir'])

        ''' SEARCH DATABASE (SQLite3) '''
        print('Building search database...')
        sqln = f"{cfg['baseName']}.sql"
        if os.path.isfile(sqln):
            with open(sqln) as sqlf:
                sql_start_db = ''.join(sqlf.readlines())
        else:
            print('Error: SQL schema file %s missing! Site search db not built.' % sqln)
            sys.exit(1)
        conn = sqlite3.connect(f"{cfg['baseName']}.db")
        c = conn.cursor()
        c.executescript(sql_start_db)
        conn.commit()

        for obj in tqdm(jsonDict):
            article = ' '.join([jsonDict[obj]['lines'][l]['text'] for l in jsonDict[obj]['lines']])
            c.execute('''insert into site_search
                values ( ?, ?, ?, ?, ?, ?, ?, ? )''', (
                obj,
                jsonDict[obj]['title'],
                article,
                jsonDict[obj]['date'],
                jsonDict[obj]['json_file'],
                jsonDict[obj]['json_dir'],
                jsonDict[obj]['json_mod_date'],
                jsonDict[obj]['slug']
                )
            )
            conn.commit()
        conn.close()
        shutil.copy2(f"{cfg['baseName']}.db", f"{cfg['baseName']}.db-backup")

        print('Build tasks complete.')

    except KeyboardInterrupt:
        sys.exit(0)
